"""
Example of using Uplink with aiohttp for non-blocking HTTP requests.
This should work on Python 3.7 and above.
"""

import asyncio

# Local imports
from github import BASE_URL, GitHub

import uplink


async def get_contributors(full_name):
    print(f"Getting GitHub repository `{full_name}`")
    response = await gh_async.get_contributors(*full_name.split("/"))
    json = await response.json()
    print(f"response for {full_name}: {json}")
    return json


if __name__ == "__main__":
    # This consumer instance uses Requests to make blocking requests.
    gh_sync = GitHub(base_url=BASE_URL)

    # This uses aiohttp, an HTTP client for asyncio.
    gh_async = GitHub(base_url=BASE_URL, client=uplink.AiohttpClient())

    # First, let's fetch a list of all public repositories.
    repos = gh_sync.get_repos().json()

    # Use only the first 10 results to avoid hitting the rate limit.
    repos = repos[:10]

    # Concurrently fetch the contributors for those repositories.
    futures = [get_contributors(repo["full_name"]) for repo in repos]
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(*futures))



from Server import BASE_URL, Google


def test_homepage():
    google = Google(base_url=BASE_URL)
    google.homepage()


def test_error():
    google = Google(base_url="NON_EXISTENT_URL")
    google.bad_page()


test_homepage()
test_error()



import uplink

# Constants
BASE_URL = "https://api.github.com/"


@uplink.headers({"Accept": "application/vnd.github.v3.full+json"})
class GitHub(uplink.Consumer):
    @uplink.get("/repositories")
    def get_repos(self):
        """Gets all public repositories."""

    @uplink.get("/repos/{owner}/{repo}/contributors")
    def get_contributors(self, owner, repo):
        """Lists contributors for the specified repository."""


if __name__ == "__main__":
    # Executes the code above
    # The for loop is used for readability, you can just call print(response.json())

    # Gets all public repositories
    github = GitHub(BASE_URL)
    response = github.get_repos()
    for repos in response.json():
        print(repos)

    # Lists contributors for a repository
    github = GitHub(BASE_URL)
    response = github.get_contributors("prkumar", "uplink")
    for contributors in response.json():
        print(contributors)



"""
This module defines a converter that uses :py:mod:`pydantic.v1` models
to deserialize and serialize values.
"""

from typing import Any

from uplink.converters.interfaces import Converter


def _encode_pydantic_v1(obj: Any) -> Any:
    from pydantic.v1.json import pydantic_encoder

    # json atoms
    if isinstance(obj, str | int | float | bool) or obj is None:
        return obj

    # json containers
    if isinstance(obj, dict):
        return {_encode_pydantic_v1(k): _encode_pydantic_v1(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_encode_pydantic_v1(i) for i in obj]

    # pydantic v1 types
    return _encode_pydantic_v1(pydantic_encoder(obj))


class _PydanticV1RequestBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, value):
        if isinstance(value, self._model):
            return _encode_pydantic_v1(value)
        return _encode_pydantic_v1(self._model.parse_obj(value))


class _PydanticV1ResponseBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, response):
        try:
            data = response.json()
        except AttributeError:
            data = response

        return self._model.parse_obj(data)



"""
This module defines a converter that uses :py:mod:`pydantic` models
to deserialize and serialize values.
"""

from typing import TYPE_CHECKING, Any

from uplink.converters.interfaces import Converter

if TYPE_CHECKING:
    # To allow using string anotations, as we wanna lazy-import
    import pydantic


def _encode_pydantic_v2(model: "pydantic.BaseModel") -> dict[str, Any]:
    return model.model_dump(mode="json")


class _PydanticV2RequestBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, value):
        if isinstance(value, self._model):
            return _encode_pydantic_v2(value)
        return _encode_pydantic_v2(self._model.model_validate(value))


class _PydanticV2ResponseBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, response):
        try:
            data = response.json()
        except AttributeError:
            data = response

        return self._model.parse_obj(data)



"""
This module is the single source of truth for any package metadata
that is used both in distribution (i.e., setup.py) and within the
codebase.
"""

import importlib.metadata

__version__ = importlib.metadata.version("uplink")



import sys

# add uplink directory to path
sys.path.insert(0, "../../")
from uplink import Consumer, error_handler, get, headers, response_handler

BASE_URL = "https://www.google.com"


def print_status(response):
    print(f"Google response status:{response.status_code}")
    return response


def handle_error(exc_type, exc_val, exc_tb):
    print(f"Error encountered. Exception will be raised. Exception Type:{exc_type}")


@headers({"Accept": "text/html"})
class Google(Consumer):
    @response_handler(print_status)
    @get("/")
    def homepage(self):
        print("google home page")
        pass

    @error_handler(handle_error)
    @get("/bad_page.html")
    def bad_page(selfs):
        pass



import pprint

import requests


def test_repo():
    r = requests.get("http://localhost:5000/repos?keyword=natural+language+processing")
    pprint.pprint(r.json())


def test_users_for_repos():
    r = requests.get(
        "http://localhost:5000/users/JustFollowUs/repo/Natural-Language-Processing"
    )
    pprint.pprint(r.json())


def test_users_for_keyword():
    r = requests.get("http://localhost:5000/users?keyword=lstm")
    pprint.pprint(r.json())


test_repo()
test_users_for_repos()
test_users_for_keyword()



import sys

# add uplink directory to path
sys.path.insert(0, "../../")
import asyncio
import os
from datetime import datetime, timedelta

from flask import Flask, jsonify, request

import uplink
from uplink import Consumer, Query, get, headers

app = Flask(__name__)

BASE_URL = "https://api.github.com"
CLIENT_ID = os.environ["CLIENT_ID"]
CLIENT_SECRET = os.environ["CLIENT_SECRET"]


@headers({"Accept": "application/vnd.github.v3+json"})
class Github(Consumer):
    @get("/search/repositories?q={keyword} in:name,description,readme")
    def repos_for_keyword(
        self,
        keyword,
        client_id: Query = CLIENT_ID,
        client_secret: Query = CLIENT_SECRET,
    ):
        """Get a list of repositories which have a given keyword in the name, description or readme"""
        pass

    @get("/repos/{user}/{repo_name}/commits")
    def commits_for_repo(
        self,
        user,
        repo_name,
        since: Query,
        client_id: Query = CLIENT_ID,
        client_secret: Query = CLIENT_SECRET,
    ):
        """Get a list of commits in a repo since some start date"""
        pass


github = Github(BASE_URL, client=uplink.AiohttpClient())
loop = asyncio.get_event_loop()

# Helpers


async def _repos_for_keyword(keyword):
    """Get repos which match the keyword search"""
    r = await github.repos_for_keyword(keyword)
    r_json = await r.json()
    return [item["full_name"] for item in r_json["items"]]


async def _users_for_repo(user, repo_name, oldest_age=55):
    """Returns users that have commited in a repo in the last N weeks"""

    since = (datetime.now() - timedelta(weeks=oldest_age)).isoformat()
    r = await github.commits_for_repo(user, repo_name, since=since)
    r_json = await r.json()
    users = set()
    for commit in r_json:
        if "author" in commit and commit["author"] is not None:
            user = (
                commit["author"]["login"],
                commit["commit"]["author"]["email"],
                commit["commit"]["author"]["name"],
            )
            users.add(user)
    return list(users)


# Flask routes


@app.route("/repos", methods=["GET"])
def repos_for_keyword():
    """
    /repos?keyword=<keyword>

    Finds all repos which contain the given keyword in the name, readme, or description"""
    if "keyword" not in request.args:
        return "", 400

    keyword = request.args["keyword"]
    future = _repos_for_keyword(keyword)
    repos = loop.run_until_complete(future)
    return jsonify(repos)


@app.route("/users/<user>/repo/<repo_name>", methods=["GET"])
def users_for_repo(user, repo_name):
    """
    /users/<user>/repo/<repo_name>[?oldest-age=<age in weeks>]

    Returns list of users who have commited in the resource user/repo in the last given amount of
    weeks"""

    oldest_age = 55 if "oldest-age" not in request.args else request.args["oldest-age"]
    future = _users_for_repo(user, repo_name, oldest_age=oldest_age)
    users = loop.run_until_complete(future)
    return jsonify(users)


@app.route("/users", methods=["GET"])
def users_for_keyword():
    """
    /users?keyword=<keyword>[?oldest-age=<age in weeks>]

    Find the top users who have commited in repositories matching the keyword in the last month"""
    if "keyword" not in request.args:
        return "", 400

    keyword = request.args["keyword"]
    oldest_age = 55 if "oldest-age" not in request.args else request.args["oldest-age"]

    repos_future = _repos_for_keyword(keyword)
    repos = loop.run_until_complete(repos_future)

    # gather futures for getting users from each repo
    users_futures = []
    users = set()
    for repo in repos:
        user, repo_name = repo.split("/")
        users_futures.append(_users_for_repo(user, repo_name, oldest_age=oldest_age))

    # barrier on all the users futures
    users_results = loop.run_until_complete(asyncio.wait(users_futures))

    # gather the results
    for users_result in users_results:
        for task in users_result:
            if task.result():
                users.update(set(task.result()))

    return jsonify(list(users))


app.run("0.0.0.0")



# Standard library imports
from pprint import pformat

# Local imports
import github

BASE_URL = "https://api.github.com/"


if __name__ == "__main__":
    # Create a GitHub API client
    gh = github.GitHub(base_url=BASE_URL)

    # Get all public repositories
    repos = gh.get_repos()

    # Shorten to first 10 results to avoid hitting the rate limit.
    repos = repos[:10]

    # Print contributors for those repositories
    for repo in repos:
        contributors = gh.get_contributors(repo.owner, repo.name)
        print(f"Contributors for {repo}:\n{pformat(contributors, indent=4)}\n")



# Local imports
from schemas import ContributorSchema, RepoSchema

from uplink import Consumer, get, headers


@headers({"Accept": "application/vnd.github.v3.full+json"})
class GitHub(Consumer):
    @get("/repositories")
    def get_repos(self) -> RepoSchema(many=True):
        """Lists all public repositories."""

    @get("/repos/{owner}/{repo}/contributors")
    def get_contributors(self, owner, repo) -> ContributorSchema(many=True):
        """Lists contributors for the specified repository."""



import pydantic

from uplink.converters import register_default_converter_factory
from uplink.converters.interfaces import Factory
from uplink.utils import is_subclass

from .pydantic_v1 import _PydanticV1RequestBody, _PydanticV1ResponseBody
from .pydantic_v2 import _PydanticV2RequestBody, _PydanticV2ResponseBody


class PydanticConverter(Factory):
    """
    A converter that serializes and deserializes values using
    :py:mod:`pydantic.v1` and `pydantic` models.

    To deserialize JSON responses into Python objects with this
    converter, define a :py:class:`pydantic.v1.BaseModel` or `pydantic.BaseModel` subclass and set
    it as the return annotation of a consumer method:

    .. code-block:: python

        @returns.json()
        @get("/users")
        def get_users(self, username) -> List[UserModel]:
            '''Fetch multiple users'''

    Note:

        This converter is an optional feature and requires the
        :py:mod:`pydantic` package. For example, here's how to
        install this feature using pip::

            $ pip install uplink[pydantic]
    """

    try:
        import pydantic
        import pydantic.v1 as pydantic_v1
    except ImportError:  # pragma: no cover
        pydantic = None
        pydantic_v1 = None

    def __init__(self):
        """
        Validates if :py:mod:`pydantic` is installed
        """
        if (self.pydantic or self.pydantic_v1) is None:
            raise ImportError("No module named 'pydantic'")

    def _get_model(self, type_):
        if is_subclass(type_, (self.pydantic_v1.BaseModel, self.pydantic.BaseModel)):
            return type_
        raise ValueError(
            "Expected pydantic.BaseModel or pydantic.v1.BaseModel subclass or instance"
        )

    def _make_converter(self, converter, type_):
        try:
            model = self._get_model(type_)
        except ValueError:
            return None

        return converter(model)

    def create_request_body_converter(self, type_, *args, **kwargs):
        if is_subclass(type_, pydantic.BaseModel):
            return self._make_converter(_PydanticV2RequestBody, type_)
        return self._make_converter(_PydanticV1RequestBody, type_)

    def create_response_body_converter(self, type_, *args, **kwargs):
        if is_subclass(type_, pydantic.BaseModel):
            return self._make_converter(_PydanticV2ResponseBody, type_)
        return self._make_converter(_PydanticV1ResponseBody, type_)

    @classmethod
    def register_if_necessary(cls, register_func):
        if cls.pydantic is not None:
            register_func(cls)


PydanticConverter.register_if_necessary(register_default_converter_factory)



# Standard library imports
import argparse
import os
import re


def is_appropriate_tag(version, tag):
    # Make sure the tag version and version in __about__.py match
    return (
        re.match(
            r"^" + tag + r"(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?$", "v" + version
        )
        is not None
    )


def is_canonical(version):
    return (
        re.match(
            r"^([1-9]\d*!)?(0|[1-9]\d*)(\.(0|[1-9]\d*))*((a|b|rc)(0|[1-9]\d*))?(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?$",
            version,
        )
        is not None
    )


def _get_current_version():
    about = {}
    with open(os.path.join("uplink", "__about__.py")) as fp:
        exec(fp.read(), about)
    return about.get("__version__", None)


def verify_version(tag):
    # Get version defined in package
    version = _get_current_version()
    assert version is not None, "The version is not defined in uplink/__about__.py."
    assert tag is not None, "The tag is not defined."
    assert is_canonical(version), "The version string [%s] violates PEP-440"
    assert is_appropriate_tag(version, tag), (
        f"The tag [{tag}] does not match the current version in uplink/__about__.py [{version}]"
    )
    return version


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", required=True)
    args = parser.parse_args()
    return verify_version(args.tag)


if __name__ == "__main__":
    print(main())



"""
Example of using Uplink with Twisted for non-blocking HTTP requests.
This should work on Python 2.7 and 3.3+.
"""

# Local imports
from github import BASE_URL, GitHub
from twisted.internet import defer, reactor

import uplink


@defer.inlineCallbacks
def get_contributors(full_name):
    print(f"Getting GitHub repository `{full_name}`")
    response = yield gh_async.get_contributors(*full_name.split("/"))
    json = response.json()
    print(f"response for {full_name}: {json}")


if __name__ == "__main__":
    # This consumer instance uses Requests to make blocking requests.
    gh_sync = GitHub(base_url=BASE_URL)

    # This uses Twisted with Requests, inspired by `requests-threads`.
    gh_async = GitHub(base_url=BASE_URL, client=uplink.TwistedClient())

    # First, let's fetch a list of all public repositories.
    repos = gh_sync.get_repos().json()

    # Use only the first 10 results to avoid hitting the rate limit.
    repos = repos[:10]

    # Asynchronously fetch the contributors for those 10 repositories.
    deferred = [get_contributors(repo["full_name"]) for repo in repos]
    reactor.callLater(2, reactor.stop)  # Stop the reactor after 2 secs
    reactor.run()



import collections

import marshmallow

# == Models == #
Contributor = collections.namedtuple(
    "Contributor", field_names=["username", "contributions"]
)

Repo = collections.namedtuple("Repo", field_names=["owner", "name"])


class SchemaBase(marshmallow.Schema):
    class Meta:
        # Pass EXCLUDE as Meta option to keep marshmallow 2 behavior
        # ref: https://marshmallow.readthedocs.io/en/3.0/upgrading.html
        unknown = getattr(marshmallow, "EXCLUDE", None)


# == Schemas == #
class ContributorSchema(SchemaBase):
    login = marshmallow.fields.Str(attribute="username")
    contributions = marshmallow.fields.Int()

    @marshmallow.post_load
    def make_contributor(self, data):
        return Contributor(**data)


class RepoSchema(SchemaBase):
    full_name = marshmallow.fields.Str()

    @marshmallow.post_load
    def make_repo(self, data):
        return Repo(*data["full_name"].split("/"))
